@@ -33,6 +33,7 @@
 
 package org.jruby;
 
+import java.util.HashSet;
 import java.util.Map;
 import java.util.Set;
 
@@ -44,8 +45,10 @@ import org.jruby.exceptions.RaiseException;
 import org.jruby.internal.runtime.methods.DynamicMethod;
 import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
+import org.jruby.runtime.CallSite;
 import org.jruby.runtime.ClassIndex;
 import org.jruby.runtime.Helpers;
+import org.jruby.runtime.JavaSites.StructSites;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.Visibility;
@@ -53,12 +56,15 @@ import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.marshal.MarshalStream;
 import org.jruby.runtime.marshal.UnmarshalStream;
 import org.jruby.util.ByteList;
+import org.jruby.util.RecursiveComparator;
 
 import static org.jruby.RubyEnumerator.enumeratorizeWithSize;
 import static org.jruby.runtime.Helpers.invokedynamic;
+import static org.jruby.runtime.ThreadContext.hasKeywords;
 import static org.jruby.runtime.Visibility.PRIVATE;
 import static org.jruby.runtime.invokedynamic.MethodNames.HASH;
 import static org.jruby.RubyEnumerator.SizeFn;
+import static org.jruby.util.RubyStringBuilder.str;
 
 /**
  * @author  jpetersen
@@ -128,6 +134,37 @@ public class RubyStruct extends RubyObject {
         testFrozen();
     }
 
+    @JRubyMethod
+    public IRubyObject deconstruct_keys(ThreadContext context, IRubyObject keysArg) {
+        if (keysArg.isNil()) return to_h(context, Block.NULL_BLOCK);
+
+        if (!(keysArg instanceof RubyArray)) {
+            throw context.runtime.newTypeError(str(context.runtime, "wrong argument type ", keysArg.getMetaClass(), "(expected Array or nil)"));
+        }
+
+        RubyArray keys = (RubyArray) keysArg;
+        int length = keys.size();
+        RubyHash hash = RubyHash.newSmallHash(context.runtime);
+        if (values.length < length) {
+            return hash;
+        }
+
+        for (int i = 0; i < length; i++) {
+            IRubyObject key = keys.eltOk(i);
+            IRubyObject value;
+
+            try {
+                value = this.aref(key);
+            } catch (RaiseException e) {
+                return hash;
+            }
+
+            hash.op_aset(context, key, value);
+        }
+
+        return hash;
+    }
+
     @JRubyMethod
     public RubyFixnum hash(ThreadContext context) {
         Ruby runtime = context.runtime;
@@ -137,7 +174,9 @@ public class RubyStruct extends RubyObject {
         IRubyObject[] values = this.values;
         for (int i = 0; i < values.length; i++) {
             h = (h << 1) | (h < 0 ? 1 : 0);
-            IRubyObject hash = context.safeRecurse(HashRecursive.INSTANCE, runtime, values[i], "hash", true);
+            IRubyObject hash = context.safeRecurse(
+                    (ctx, runtime1, obj, recur) -> recur ? RubyFixnum.zero(runtime1) : invokedynamic(ctx, obj, HASH),
+                    runtime, values[i], "hash", true);
             h ^= RubyNumeric.num2long(hash);
         }
 
@@ -177,14 +216,16 @@ public class RubyStruct extends RubyObject {
      * MRI: rb_struct_s_def / make_struct
      *
      */
-    @JRubyMethod(name = "new", required = 1, rest = true, meta = true)
+    @JRubyMethod(name = "new", required = 1, rest = true, checkArity = false, meta = true, keywords = true)
     public static RubyClass newInstance(IRubyObject recv, IRubyObject[] args, Block block) {
-        String name = null;
-        boolean nilName = false;
-        boolean keywordInit = false;
         Ruby runtime = recv.getRuntime();
 
-        if (args.length > 0) {
+        int argc = Arity.checkArgumentCount(runtime, args, 1, -1);
+
+        String name = null;
+        boolean nilName = false;
+
+        if (argc > 0) {
             IRubyObject firstArgAsString = args[0].checkStringType();
             if (!firstArgAsString.isNil()) {
                 RubySymbol nameSym = ((RubyString)firstArgAsString).intern();
@@ -198,15 +239,15 @@ public class RubyStruct extends RubyObject {
         }
 
         RubyArray member = runtime.newArray();
+        IRubyObject keywordInitValue = runtime.getNil();
 
-        int argc = args.length;
-        final IRubyObject opts = args[args.length - 1];
+        final IRubyObject opts = args[argc - 1];
         if (opts instanceof RubyHash) {
             argc--;
-            IRubyObject ret = ArgsUtil.extractKeywordArg(runtime.getCurrentContext(), (RubyHash) opts, "keyword_init");
-            keywordInit = ret != null && ret.isTrue();
+            keywordInitValue = ArgsUtil.extractKeywordArg(runtime.getCurrentContext(), (RubyHash) opts, "keyword_init");
         }
 
+        Set<IRubyObject> tmpMemberSet = new HashSet<IRubyObject>();
         for (int i = (name == null && !nilName) ? 0 : 1; i < argc; i++) {
             IRubyObject arg = args[i];
             RubySymbol sym;
@@ -217,6 +258,11 @@ public class RubyStruct extends RubyObject {
             } else {
                 sym = runtime.newSymbol(arg.asJavaString());
             }
+            if (tmpMemberSet.contains(sym)) {
+                throw runtime.newArgumentError("duplicate member: " + sym);
+            } else {
+                tmpMemberSet.add(sym);
+            }
 
             member.append(sym);
         }
@@ -245,13 +291,13 @@ public class RubyStruct extends RubyObject {
 
         newStruct.setInternalVariable(SIZE_VAR, member.length());
         newStruct.setInternalVariable(MEMBER_VAR, member);
-        newStruct.setInternalVariable(KEYWORD_INIT_VAR, keywordInit ? runtime.getTrue() : runtime.getFalse());
+        newStruct.setInternalVariable(KEYWORD_INIT_VAR, keywordInitValue);
 
         newStruct.getSingletonClass().defineAnnotatedMethods(StructMethods.class);
 
         // define access methods.
-        for (int i = (name == null && !nilName) ? 0 : 1; i < args.length; i++) {
-            if (i == args.length - 1 && args[i] instanceof RubyHash) break;
+        for (int i = (name == null && !nilName) ? 0 : 1; i < argc; i++) {
+            if (i == argc - 1 && args[i] instanceof RubyHash) break;
             final String memberName = args[i].asJavaString();
             // if we are storing a name as well, index is one too high for values
             final int index = (name == null && !nilName) ? i : i - 1;
@@ -273,27 +319,27 @@ public class RubyStruct extends RubyObject {
 
     // For binding purposes on the newly created struct types
     public static class StructMethods {
-        @JRubyMethod(name = {"new", "[]"}, rest = true)
+        @JRubyMethod(name = {"new", "[]"}, rest = true, keywords = true)
         public static IRubyObject newStruct(IRubyObject recv, IRubyObject[] args, Block block) {
             return RubyStruct.newStruct(recv, args, block);
         }
 
-        @JRubyMethod(name = {"new", "[]"})
+        @JRubyMethod(name = {"new", "[]"}, keywords = true)
         public static IRubyObject newStruct(IRubyObject recv, Block block) {
             return RubyStruct.newStruct(recv, block);
         }
 
-        @JRubyMethod(name = {"new", "[]"})
+        @JRubyMethod(name = {"new", "[]"}, keywords = true)
         public static IRubyObject newStruct(IRubyObject recv, IRubyObject arg0, Block block) {
             return RubyStruct.newStruct(recv, arg0, block);
         }
 
-        @JRubyMethod(name = {"new", "[]"})
+        @JRubyMethod(name = {"new", "[]"}, keywords = true)
         public static IRubyObject newStruct(IRubyObject recv, IRubyObject arg0, IRubyObject arg1, Block block) {
             return RubyStruct.newStruct(recv, arg0, arg1, block);
         }
 
-        @JRubyMethod(name = {"new", "[]"})
+        @JRubyMethod(name = {"new", "[]"}, keywords = true)
         public static IRubyObject newStruct(IRubyObject recv, IRubyObject arg0, IRubyObject arg1, IRubyObject arg2, Block block) {
             return RubyStruct.newStruct(recv, arg0, arg1, arg2, block);
         }
@@ -309,6 +355,11 @@ public class RubyStruct extends RubyObject {
             if (!keywordInit.isTrue()) return recv.inspect();
             return recv.inspect().convertToString().catString("(keyword_init: true)");
         }
+
+        @JRubyMethod(name = "keyword_init?")
+        public static IRubyObject keyword_init_p(IRubyObject self) {
+            return RubyStruct.getInternalVariable((RubyClass) self, KEYWORD_INIT_VAR);
+        }
     }
 
     /** Create new Structure.
@@ -362,13 +413,20 @@ public class RubyStruct extends RubyObject {
         }
     }
 
-    @JRubyMethod(rest = true, visibility = PRIVATE)
+    private void checkForKeywords(ThreadContext context, boolean keywordInit) {
+        if (hasKeywords(ThreadContext.resetCallInfo(context)) && !keywordInit) {
+            context.runtime.getWarnings().warn("Passing only keyword arguments to Struct#initialize will behave differently from Ruby 3.2. Please use a Hash literal like .new({k: v}) instead of .new(k: v).");
+        }
+    }
+
+    @JRubyMethod(rest = true, visibility = PRIVATE, keywords = true)
     public IRubyObject initialize(ThreadContext context, IRubyObject[] args) {
+        IRubyObject keywordInit = RubyStruct.getInternalVariable(classOf(), KEYWORD_INIT_VAR);
+        checkForKeywords(context, !keywordInit.isNil());
+        ThreadContext.resetCallInfo(context);
         modify();
         checkSize(args.length);
 
-        IRubyObject keywordInit = RubyStruct.getInternalVariable(classOf(), KEYWORD_INIT_VAR);
-
         if (keywordInit.isTrue()) {
             if (args.length != 1) throw context.runtime.newArgumentError(args.length, 0);
 
@@ -382,6 +440,8 @@ public class RubyStruct extends RubyObject {
     }
 
     private void setupStructValuesFromHash(ThreadContext context, RubyHash kwArgs) {
+        Ruby runtime = context.runtime;
+
         RubyArray __members__ = __member__();
         Set<Map.Entry<IRubyObject, IRubyObject>> entries = kwArgs.directEntrySet();
 
@@ -389,9 +449,9 @@ public class RubyStruct extends RubyObject {
                 entry -> {
                     IRubyObject key = entry.getKey();
                     if (!(key instanceof RubySymbol))
-                        key = context.runtime.newSymbol(key.convertToString().getByteList());
+                        key = runtime.newSymbol(key.convertToString().getByteList());
                     IRubyObject index = __members__.index(context, key);
-                    if (index.isNil()) throw context.runtime.newArgumentError("unknown keywords: " + key);
+                    if (index.isNil()) throw runtime.newArgumentError(str(runtime, "unknown keywords: ", key));
                     values[index.convertToInteger().getIntValue()] = entry.getValue();
                 });
     }
@@ -405,9 +465,11 @@ public class RubyStruct extends RubyObject {
 
     @JRubyMethod(visibility = PRIVATE)
     public IRubyObject initialize(ThreadContext context, IRubyObject arg0) {
+        IRubyObject keywordInit = RubyStruct.getInternalVariable(classOf(), KEYWORD_INIT_VAR);
+        checkForKeywords(context, !keywordInit.isNil());
+        ThreadContext.resetCallInfo(context);
         Ruby runtime = context.runtime;
 
-        IRubyObject keywordInit = RubyStruct.getInternalVariable(classOf(), KEYWORD_INIT_VAR);
         if (keywordInit.isTrue()) {
             IRubyObject maybeKwargs = ArgsUtil.getOptionsArg(runtime, arg0);
 
@@ -546,7 +608,7 @@ public class RubyStruct extends RubyObject {
         System.arraycopy(values, 0, struct.values, 0, values.length);
     }
 
-    @JRubyMethod(name = "==", required = 1)
+    @JRubyMethod(name = "==")
     @Override
     public IRubyObject op_equal(ThreadContext context, IRubyObject other) {
         if (this == other) return context.tru;
@@ -557,11 +619,10 @@ public class RubyStruct extends RubyObject {
 
         if (other == this) return context.tru;
 
-        // recursion guard
-        return context.safeRecurse(EqualRecursive.INSTANCE, other, this, "==", true);
+        return RecursiveComparator.compare(context, sites(context).op_equal, this, other);
     }
 
-    @JRubyMethod(name = "eql?", required = 1)
+    @JRubyMethod(name = "eql?")
     public IRubyObject eql_p(ThreadContext context, IRubyObject other) {
         if (this == other) return context.tru;
         if (!(other instanceof RubyStruct)) return context.fals;
@@ -569,8 +630,26 @@ public class RubyStruct extends RubyObject {
             return context.fals;
         }
 
-        // recursion guard
-        return context.safeRecurse(EqlRecursive.INSTANCE, other, this, "eql?", true);
+        return RecursiveComparator.compare(context, sites(context).eql, this, other);
+    }
+
+    public RubyBoolean compare(ThreadContext context, CallSite site, IRubyObject other) {
+        if (!(other instanceof RubyStruct)) {
+            return context.fals;
+        }
+
+        RubyStruct struct = (RubyStruct) other;
+
+        if (values.length != struct.values.length) return context.fals;
+
+        for (int i = 0; i < values.length; i++) {
+            IRubyObject a = aref(i);
+            IRubyObject b = struct.aref(i);
+
+            if (!site.call(context, a, a, b).isTrue()) return context.fals;
+        }
+
+        return context.tru;
     }
 
     private static final byte[] STRUCT_BEG = { '#','<','s','t','r','u','c','t',' ' };
@@ -613,16 +692,16 @@ public class RubyStruct extends RubyObject {
         }
 
         buffer.cat('>');
-        return (RubyString) buffer.infectBy(this);
+        return (RubyString) buffer;
     }
 
     @JRubyMethod(name = {"inspect", "to_s"})
     public RubyString inspect(final ThreadContext context) {
         // recursion guard
-        return (RubyString) context.safeRecurse(InspectRecursive.INSTANCE, this, this, "inspect", false);
+        return (RubyString) context.safeRecurse((ctx, self, obj, recur) -> self.inspectStruct(ctx, recur), this, this, "inspect", false);
     }
 
-    @JRubyMethod(name = {"to_a", "values"})
+    @JRubyMethod(name = {"to_a", "deconstruct", "values"})
     @Override
     public RubyArray to_a(ThreadContext context) {
         return context.runtime.newArray(values);
@@ -697,7 +776,7 @@ public class RubyStruct extends RubyObject {
         return block.isGiven() ? each_pairInternal(context, block) : enumeratorizeWithSize(context, this, "each_pair", RubyStruct::size);
     }
 
-    @JRubyMethod(name = "[]", required = 1)
+    @JRubyMethod(name = "[]")
     public IRubyObject aref(IRubyObject key) {
         return arefImpl( key, false );
     }
@@ -728,7 +807,7 @@ public class RubyStruct extends RubyObject {
         return values[newIdx];
     }
 
-    @JRubyMethod(name = "[]=", required = 2)
+    @JRubyMethod(name = "[]=")
     public IRubyObject aset(IRubyObject key, IRubyObject value) {
         if (key instanceof RubyString || key instanceof RubySymbol) {
             final String name = key.asJavaString();
@@ -753,35 +832,37 @@ public class RubyStruct extends RubyObject {
         return values[newIdx] = value;
     }
 
-    // NOTE: copied from RubyArray, both RE, Struct, and Array should share one impl
     @JRubyMethod(rest = true)
     public IRubyObject values_at(IRubyObject[] args) {
+        final Ruby runtime = metaClass.runtime;
         final int olen = values.length;
         RubyArray result = getRuntime().newArray(args.length);
 
         for (int i = 0; i < args.length; i++) {
             final IRubyObject arg = args[i];
-            if ( arg instanceof RubyFixnum ) {
+            if (arg instanceof RubyFixnum ) {
                 result.append( aref(arg) );
                 continue;
             }
 
-            final int[] begLen;
-            if ( ! ( arg instanceof RubyRange ) ) {
-                // do result.append
-            }
-            else if ( ( begLen = ((RubyRange) args[i]).begLenInt(olen, 0) ) == null ) {
-                continue;
-            }
-            else {
+            final int [] begLen = new int[2];
+            if (arg instanceof RubyRange &&
+                    RubyRange.rangeBeginLength(runtime.getCurrentContext(),(RubyRange) args[i], olen, begLen, 1).isTrue()) {
                 final int beg = begLen[0];
                 final int len = begLen[1];
-                for (int j = 0; j < len; j++) {
-                    result.append( aref(j + beg) );
+                int end = olen < beg + len ? olen : beg + len;
+                int j;
+                for (j = beg; j < end; j++) {
+                    result.push(aref(j));
+                }
+                if ( beg + len > j ) {
+                    IRubyObject [] tmp = new IRubyObject[beg + len - j];
+                    Helpers.fillNil(tmp, getRuntime());
+                    result.push(tmp);
                 }
                 continue;
             }
-            result.append( aref(RubyNumeric.num2int(arg)) );
+            result.push(aref(RubyNumeric.num2int(arg)));
         }
 
         return result;
@@ -804,10 +885,12 @@ public class RubyStruct extends RubyObject {
         return RubyObject.dig2(context, val, arg1, arg2);
     }
 
-    @JRubyMethod(name = "dig", required = 1, rest = true)
+    @JRubyMethod(name = "dig", required = 1, rest = true, checkArity = false)
     public IRubyObject dig(ThreadContext context, IRubyObject[] args) {
+        int argc = Arity.checkArgumentCount(context, args, 1, -1);
+
         final IRubyObject val = arefImpl( args[0], true );
-        return args.length == 1 ? val : RubyObject.dig(context, val, args, 1);
+        return argc == 1 ? val : RubyObject.dig(context, val, args, 1);
     }
 
     public static void marshalTo(RubyStruct struct, MarshalStream output) throws java.io.IOException {
@@ -827,7 +910,7 @@ public class RubyStruct extends RubyObject {
     public static RubyStruct unmarshalFrom(UnmarshalStream input) throws java.io.IOException {
         final Ruby runtime = input.getRuntime();
 
-        RubySymbol className = (RubySymbol) input.unmarshalObject(false);
+        RubySymbol className = input.unique();
         RubyClass rbClass = pathToClass(runtime, className.asJavaString());
         if (rbClass == null) {
             throw runtime.newNameError(UNINITIALIZED_CONSTANT, runtime.getStructClass(), className);
@@ -838,16 +921,14 @@ public class RubyStruct extends RubyObject {
         final int len = input.unmarshalInt();
 
         // FIXME: This could all be more efficient, but it's how struct works
-        final RubyStruct result;
         // 1.9 does not appear to call initialize (JRUBY-5875)
-        result = new RubyStruct(runtime, rbClass);
-        input.registerLinkTarget(result);
+        final RubyStruct result = (RubyStruct) input.entry(new RubyStruct(runtime, rbClass));
 
         for (int i = 0; i < len; i++) {
-            IRubyObject slot = input.unmarshalObject(false);
-            final IRubyObject elem = member.eltInternal(i); // RubySymbol
-            if ( ! elem.toString().equals( slot.toString() ) ) {
-                throw runtime.newTypeError("struct " + rbClass.getName() + " not compatible (:" + slot + " for :" + elem + ")");
+            RubySymbol slot = input.symbol();
+            RubySymbol elem = (RubySymbol) member.eltInternal(i);
+            if (!elem.equals(slot)) {
+                throw runtime.newTypeError(str(runtime, "struct ", rbClass, " not compatible (:", slot, " for :", elem, ")"));
             }
             result.aset(i, input.unmarshalObject());
         }
@@ -861,7 +942,7 @@ public class RubyStruct extends RubyObject {
     }
 
     @Override
-    @JRubyMethod(required = 1, visibility = Visibility.PRIVATE)
+    @JRubyMethod(visibility = Visibility.PRIVATE)
     public IRubyObject initialize_copy(IRubyObject arg) {
         if (this == arg) return this;
         RubyStruct original = (RubyStruct) arg;
@@ -953,17 +1034,6 @@ public class RubyStruct extends RubyObject {
         }
     }
 
-    private static class HashRecursive implements ThreadContext.RecursiveFunctionEx<Ruby> {
-
-        static final HashRecursive INSTANCE = new HashRecursive();
-
-        @Override
-        public IRubyObject call(ThreadContext context, Ruby runtime, IRubyObject obj, boolean recur) {
-            if (recur) return RubyFixnum.zero(runtime);
-            return invokedynamic(context, obj, HASH);
-        }
-    }
-
     private static class EqualRecursive implements ThreadContext.RecursiveFunctionEx<IRubyObject> {
 
         private static final EqualRecursive INSTANCE = new EqualRecursive();
@@ -981,13 +1051,8 @@ public class RubyStruct extends RubyObject {
         }
     }
 
-    private static class InspectRecursive implements ThreadContext.RecursiveFunctionEx<RubyStruct> {
-
-        private static final ThreadContext.RecursiveFunctionEx INSTANCE = new InspectRecursive();
-
-        public IRubyObject call(ThreadContext context, RubyStruct self, IRubyObject obj, boolean recur) {
-            return self.inspectStruct(context, recur);
-        }
+    private static StructSites sites(ThreadContext context) {
+        return context.sites.Struct;
     }
 
     @Deprecated
